"""This module defines custom implementation of ``clean`` setuptools command."""

from __future__ import annotations

import os
from shutil import rmtree

import setuptools  # noqa: F401
from distutils.command.clean import clean as _clean

from ..constants import CMAKE_BUILD_DIR, CMAKE_INSTALL_DIR, SKBUILD_DIR
from ..utils import logger
from . import set_build_base_mixin


class clean(set_build_base_mixin, _clean):
    """Custom implementation of ``clean`` setuptools command."""

    def run(self) -> None:
        """After calling the super class implementation, this function removes
        the directories specific to scikit-build."""
        super().run()
        for dir_ in (CMAKE_INSTALL_DIR(), CMAKE_BUILD_DIR(), SKBUILD_DIR()):
            if os.path.exists(dir_):
                logger.info("removing '%s'", dir_)
            # This seems to be there but isn't typed in the stubs TODO
            if not self.dry_run and os.path.exists(dir_):  # type: ignore[attr-defined]
                rmtree(dir_)



"""This module defines object specific to OSX platform."""

from __future__ import annotations

import sys
import textwrap

from . import unix


class OSXPlatform(unix.UnixPlatform):
    """OSX implementation of :class:`.abstract.CMakePlatform`."""

    @property
    def generator_installation_help(self) -> str:
        """Return message guiding the user for installing a valid toolchain."""
        return (
            textwrap.dedent(
                """
            Building MacOSX wheels for Python {pyver} requires XCode.
            Get it here:

              https://developer.apple.com/xcode/
            """
            )
            .format(pyver=".".join(str(v) for v in sys.version_info[:2]))
            .strip()
        )



"""This module defines custom ``generate_source_manifest`` setuptools
command."""

from __future__ import annotations

import os
import subprocess
import sys

from distutils.cmd import Command

from ..constants import SKBUILD_DIR, SKBUILD_MARKER_FILE
from . import set_build_base_mixin


class generate_source_manifest(set_build_base_mixin, Command):
    """Custom setuptools command generating a `MANIFEST` file if
    not already provided."""

    description = "generate source MANIFEST"

    def initialize_options(self) -> None:
        """Set default values for all the options that this command supports."""

    def run(self) -> None:
        """
        If neither a `MANIFEST`, nor a `MANIFEST.in` file is provided, and
        we are in a git repo, try to create a `MANIFEST.in` file from the output of
        `git ls-tree --name-only -r HEAD`.

        We need a reliable way to tell if an existing `MANIFEST` file is one
        we've generated.  distutils already uses a first-line comment to tell
        if the `MANIFEST` file was generated from `MANIFEST.in`, so we use a
        dummy file, `_skbuild_MANIFEST`, to avoid confusing distutils.
        """
        do_generate = (
            # If there's a MANIFEST.in file, we assume that we had nothing to do
            # with the project's manifest.
            not os.path.exists("MANIFEST.in")
            # otherwise, we check to see that there is no MANIFEST, ...
            if not os.path.exists("MANIFEST")
            # ... (if there is one,) that we created it
            else os.path.exists(SKBUILD_MARKER_FILE())
        )

        if do_generate:
            try:
                with open("MANIFEST.in", "wb") as manifest_in_file:
                    # Since Git < 2.11 does not support --recurse-submodules option, fallback to
                    # regular listing.
                    try:
                        cmd_out = subprocess.run(
                            ["git", "ls-files", "--recurse-submodules"], stdout=subprocess.PIPE, check=True
                        ).stdout
                    except subprocess.CalledProcessError:
                        cmd_out = subprocess.run(["git", "ls-files"], stdout=subprocess.PIPE, check=True).stdout
                    git_files = [git_file.strip() for git_file in cmd_out.split(b"\n")]
                    manifest_text = b"\n".join([b"include %s" % git_file.strip() for git_file in git_files if git_file])
                    manifest_text += b"\nexclude MANIFEST.in"
                    manifest_in_file.write(manifest_text)
            except subprocess.CalledProcessError:
                sys.stderr.write(
                    "\n\n"
                    "Since scikit-build could not find MANIFEST.in or "
                    "MANIFEST, it tried to generate a MANIFEST.in file "
                    "automatically, but could not because it could not "
                    "determine which source files to include.\n\n"
                    'The command used was "git ls-files"\n'
                    "\n\n"
                )
                raise

            if not os.path.exists(SKBUILD_DIR()):
                os.makedirs(SKBUILD_DIR())

            with open(SKBUILD_MARKER_FILE(), "w", encoding="utf-8"):  # touch
                pass

    def finalize_options(self, *args: object, **kwargs: object) -> None:
        """Set final values for all the options that this command supports."""



"""This module defines object specific to AIX platform."""

from __future__ import annotations

import sys
import textwrap

from . import unix


class AIXPlatform(unix.UnixPlatform):
    """AIX implementation of :class:`.abstract.CMakePlatform`."""

    @property
    def generator_installation_help(self) -> str:
        """Return message guiding the user for installing a valid toolchain."""
        return (
            textwrap.dedent(
                """
            Building AIX wheels for Python {pyver} requires IBM XL C/C++.
            Get it here:

              https://www.ibm.com/products/xl-c-aix-compiler-power
            """
            )
            .format(pyver=".".join(str(v) for v in sys.version_info[:2]))
            .strip()
        )



"""This package provides :func:`get_platform()` allowing to get an instance
of :class:`.abstract.CMakePlatform` matching the current platform.

This folder contains files the define CMake's defaults for given platforms.  Any of them can be overridden by either
command line or by environment variables.
"""

from __future__ import annotations

from .abstract import CMakeGenerator
from .platform_factory import get_platform

__all__ = ["CMakeGenerator", "get_platform"]



"""This module defines functions generally useful in scikit-build."""

from __future__ import annotations

import contextlib
import logging
import os
import typing
from contextlib import contextmanager
from typing import Any, Iterable, Iterator, Mapping, NamedTuple, Sequence, TypeVar

from distutils.command.build_py import build_py as distutils_build_py
from distutils.errors import DistutilsTemplateError
from distutils.filelist import FileList
from distutils.text_file import TextFile

from .._compat.typing import Protocol

if typing.TYPE_CHECKING:
    import setuptools._distutils.dist


class CommonLog(Protocol):
    """Protocol for loggers with an info method."""

    # pylint: disable-next=missing-function-docstring
    def info(self, __msg: str, *args: object) -> None: ...


logger: CommonLog

try:
    import setuptools.logging

    skb_log = logging.getLogger("skbuild")
    skb_log.setLevel(logging.INFO)
    logging_module = True
    logger = skb_log

except ImportError:
    from distutils import log as distutils_log

    logger = distutils_log
    logging_module = False


class Distribution(NamedTuple):
    """Distribution stand-in."""

    script_name: str


def _log_warning(msg: str, *args: object) -> None:
    try:
        if logging_module:
            skb_log.warning(msg, *args)
        else:
            distutils_log.warn(msg, *args)
    except ValueError:
        # Setuptools might disconnect the logger. That shouldn't be an error for a warning.
        print(msg % args, flush=True)


def mkdir_p(path: str) -> None:
    """Ensure directory ``path`` exists. If needed, parent directories
    are created.
    """
    return os.makedirs(path, exist_ok=True)


Self = TypeVar("Self", bound="push_dir")


class push_dir(contextlib.ContextDecorator):
    """Context manager to change current directory."""

    def __init__(self, directory: str | None = None, make_directory: bool = False) -> None:
        """
        :param directory:
          Path to set as current working directory. If ``None``
          is passed, ``os.getcwd()`` is used instead.

        :param make_directory:
          If True, ``directory`` is created.
        """
        super().__init__()
        self.directory = directory
        self.make_directory = make_directory
        self.old_cwd: str | None = None

    def __enter__(self: Self) -> Self:
        self.old_cwd = os.getcwd()
        if self.directory:
            if self.make_directory:
                os.makedirs(self.directory, exist_ok=True)
            os.chdir(self.directory)
        return self

    def __exit__(self, typ: None, val: None, traceback: None) -> None:
        assert self.old_cwd is not None
        os.chdir(self.old_cwd)


class PythonModuleFinder(distutils_build_py):
    """Convenience class to search for python modules.

    This class is based on ``distutils.command.build_py.build_by`` and
    provides a specialized version of ``find_all_modules()``.
    """

    distribution: Distribution  # type: ignore[assignment]

    # pylint: disable-next=super-init-not-called
    def __init__(
        self,
        packages: Sequence[str],
        package_dir: Mapping[str, str],
        py_modules: Sequence[str],
        alternative_build_base: str | None = None,
    ) -> None:
        """
        :param packages: List of packages to search.
        :param package_dir: Dictionary mapping ``package`` with ``directory``.
        :param py_modules: List of python modules.
        :param alternative_build_base: Additional directory to search in.
        """
        self.packages = packages
        self.package_dir = package_dir
        self.py_modules = py_modules
        self.alternative_build_base = alternative_build_base

        self.distribution = Distribution("setup.py")

    def find_all_modules(self, project_dir: str | None = None) -> list[Any | tuple[str, str, str]]:
        """Compute the list of all modules that would be built by
        project located in current directory, whether they are
        specified one-module-at-a-time ``py_modules`` or by whole
        packages ``packages``.

        By default, the function will search for modules in the current
        directory. Specifying ``project_dir`` parameter allow to change
        this.

        Return a list of tuples ``(package, module, module_file)``.
        """
        with push_dir(project_dir):
            # TODO: typestubs for distutils
            return super().find_all_modules()  # type: ignore[no-any-return, no-untyped-call]

    def find_package_modules(self, package: str, package_dir: str) -> Iterable[tuple[str, str, str]]:
        """Temporally prepend the ``alternative_build_base`` to ``module_file``.
        Doing so will ensure modules can also be found in other location
        (e.g ``skbuild.constants.CMAKE_INSTALL_DIR``).
        """
        if package_dir and not os.path.exists(package_dir) and self.alternative_build_base is not None:
            package_dir = os.path.join(self.alternative_build_base, package_dir)

        modules: Iterable[tuple[str, str, str]] = super().find_package_modules(package, package_dir)  # type: ignore[no-untyped-call]

        # Strip the alternative base from module_file
        def _strip_directory(entry: tuple[str, str, str]) -> tuple[str, str, str]:
            module_file = entry[2]
            if self.alternative_build_base is not None and module_file.startswith(self.alternative_build_base):
                module_file = module_file[len(self.alternative_build_base) + 1 :]
            return entry[0], entry[1], module_file

        return map(_strip_directory, modules)

    def check_module(self, module: str, module_file: str) -> bool:
        """Return True if ``module_file`` belongs to ``module``."""
        if self.alternative_build_base is not None:
            updated_module_file = os.path.join(self.alternative_build_base, module_file)
            if os.path.exists(updated_module_file):
                module_file = updated_module_file
        if not os.path.isfile(module_file):
            _log_warning("file %s (for module %s) not found", module_file, module)
            return False
        return True


OptStr = TypeVar("OptStr", str, None)


def to_platform_path(path: OptStr) -> OptStr:
    """Return a version of ``path`` where all separator are :attr:`os.sep`"""
    if path is None:
        return path
    return path.replace("/", os.sep).replace("\\", os.sep)


def to_unix_path(path: OptStr) -> OptStr:
    """Return a version of ``path`` where all separator are ``/``"""
    if path is None:
        return path
    return path.replace("\\", "/")


@contextmanager
def distribution_hide_listing(
    distribution: setuptools._distutils.dist.Distribution | Distribution,
) -> Iterator[bool | int]:
    """Given a ``distribution``, this context manager temporarily
    sets distutils threshold to WARN if ``--hide-listing`` argument
    was provided.

    It yields True if ``--hide-listing`` argument was provided.
    """

    hide_listing = getattr(distribution, "hide_listing", False)
    wheel_log = logging.getLogger("wheel")
    root_log = logging.getLogger()  # setuptools 65.6+ needs this hidden too
    if logging_module:
        # Setuptools 60.2+, will always be on Python 3.7+
        old_wheel_level = wheel_log.getEffectiveLevel()
        old_root_level = root_log.getEffectiveLevel()
        try:
            if hide_listing:
                wheel_log.setLevel(logging.WARNING)
                root_log.setLevel(logging.WARNING)
                # The classic logger doesn't respond to set_threshold anymore,
                # but it does log info and above to stdout, so let's hide that
                with open(os.devnull, "w", encoding="utf-8") as f, contextlib.redirect_stdout(f):
                    yield hide_listing
            else:
                yield hide_listing
        finally:
            if hide_listing:
                wheel_log.setLevel(old_wheel_level)
                root_log.setLevel(old_root_level)

    else:
        old_threshold = distutils_log._global_log.threshold  # type: ignore[attr-defined]
        if hide_listing:
            distutils_log.set_threshold(distutils_log.WARN)
        try:
            yield hide_listing
        finally:
            distutils_log.set_threshold(old_threshold)


def parse_manifestin(template: str) -> list[str]:
    """This function parses template file (usually MANIFEST.in)"""
    if not os.path.exists(template):
        return []

    template_file = TextFile(
        template,
        strip_comments=True,
        skip_blanks=True,
        join_lines=True,
        lstrip_ws=True,
        rstrip_ws=True,
        collapse_join=True,
    )

    file_list = FileList()
    try:
        while True:
            line = template_file.readline()
            if line is None:  # end of file
                break

            try:
                file_list.process_template_line(line)
            # the call above can raise a DistutilsTemplateError for
            # malformed lines, or a ValueError from the lower-level
            # convert_path function
            except (DistutilsTemplateError, ValueError) as msg:
                filename = template_file.filename if hasattr(template_file, "filename") else "Unknown"
                current_line = template_file.current_line if hasattr(template_file, "current_line") else "Unknown"
                print(f"{filename}, line {current_line}: {msg}", flush=True)
        return file_list.files
    finally:
        template_file.close()
